from functools import partial

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AbstractBaseUser
from django.utils.deprecation import MiddlewareMixin
from django.utils.functional import SimpleLazyObject

from django_firebase_auth import auth

# 설정 값 읽기
AUTHORIZATION_HEADER = getattr(settings, "AUTHORIZATION_HEADER", "Authorization")

COOKIE_NAME = getattr(settings, "COOKIE_NAME", None)
COOKIE_MAX_AGE = getattr(settings, "COOKIE_MAX_AGE", 3600)
COOKIE_HTTP_ONLY = getattr(settings, "COOKIE_HTTP_ONLY", True)
COOKIE_SECURE = getattr(settings, "COOKIE_SECURE", True)
COOKIE_SAMESITE = getattr(settings, "COOKIE_SAMESITE", "Lax")


# 사용자 모델 타입 가져오기
UserModel: AbstractBaseUser = get_user_model()


def get_user(request):
    if not hasattr(request, "_cached_user"):
        request._cached_user = auth.get_user(request)
    return request._cached_user


async def auser(request):
    if not hasattr(request, "_acached_user"):
        request._acached_user = await auth.aget_user(request)
    return request._acached_user


class AuthenticationMiddleware(MiddlewareMixin):

    def process_request(self, request):
        request.user = SimpleLazyObject(lambda: get_user(request))
        request.auser = partial(auser, request)

    def process_response(self, request, response):
        token = request.headers.get(AUTHORIZATION_HEADER)
        if not token:
            return response
        if not auth.is_valid_jwt(token):
            return response

        # 쿠키에 토큰 설정
        if COOKIE_NAME:
            response.set_cookie(
                COOKIE_NAME,  # 쿠키 이름
                token,  # 토큰 값
                max_age=COOKIE_MAX_AGE,  # 쿠키 유효 시간 (초 단위, 여기서는 1시간)
                httponly=COOKIE_HTTP_ONLY,
                secure=COOKIE_SECURE,
                samesite=COOKIE_SAMESITE,
            )
        return response



import binascii
from base64 import b64decode, b64encode
from typing import NamedTuple, Optional, Union

from django.db import models
from django.forms import model_to_dict
from graphql.type.scalars import serialize_id

Base64String = str


class ResolvedGlobalId(NamedTuple):
    type: str
    id: str


def base64(s: str) -> Base64String:
    """Encode the string s using Base64."""
    b: bytes = s.encode("utf-8") if isinstance(s, str) else s
    return b64encode(b).decode("ascii")


def unbase64(s: Base64String) -> str:
    """Decode the string s using Base64."""
    try:
        b: bytes = s.encode("ascii") if isinstance(s, str) else s
    except UnicodeEncodeError:
        return ""
    try:
        return b64decode(b).decode("utf-8")
    except (binascii.Error, UnicodeDecodeError):
        return ""


def is_valid_base64(s: str) -> bool:
    try:
        return b64encode(b64decode(s)) == s.encode()
    except Exception:
        return False


def from_global_id(global_id: str) -> ResolvedGlobalId:
    """
    to_global_id로 생성된 "global ID"를 받아서
    이를 생성하는데 사용된 type name과 ID를 반환합니다.
    """
    if not is_valid_base64(global_id):
        raise ValueError("global_id가 유효하지 않습니다.")

    global_id = unbase64(global_id)
    return ResolvedGlobalId(*global_id.split(":", 1))


def to_global_id(type_: str, id_: Union[str, int]) -> str:
    """
    Takes a type name and an ID specific to that type name, and returns a
    "global ID" that is unique among all types.
    """
    return base64(f"{type_}:{serialize_id(id_)}")


class BaseORMEnum(models.TextChoices):
    @classmethod
    def of(cls, value) -> Optional["BaseORMEnum"]:
        if value is None:
            return None

        for tag in cls:
            if tag.value == value:
                return tag
        return None


class BaseORMModel(models.Model):
    created_at = models.DateTimeField(auto_now_add=True, help_text="생성일")
    modified_at = models.DateTimeField(auto_now=True, help_text="수정일")

    class Meta:
        abstract = True

    @property
    def created_ts(self) -> float:
        return self.created_at.timestamp() * 1e3

    @property
    def modified_ts(self):
        return self.modified_at.timestamp() * 1e3

    @classmethod
    def _check_global_id_type(cls, global_id: str):
        resolved_global_id = from_global_id(global_id)

        if resolved_global_id.type != cls.__name__:
            raise ValueError("global_id의 type이 일치하지 않습니다.")

        return resolved_global_id.id

    @classmethod
    def get_by_global_id(cls, global_id: str):
        _id = cls._check_global_id_type(global_id)

        return cls.objects.get(pk=_id)

    @classmethod
    def get_with_global_id(cls, global_id: str, **kwargs):
        _id = cls._check_global_id_type(global_id)

        return cls.objects.get(pk=_id, **kwargs)

    @classmethod
    def filter_by_global_id(cls, global_id: str, **kwargs):
        _id = cls._check_global_id_type(global_id)

        return cls.objects.filter(pk=_id, **kwargs)

    @classmethod
    def id_from_global_id(cls, global_id: str):
        return cls._check_global_id_type(global_id)

    @property
    def global_id(self) -> str:
        return to_global_id(self.__class__.__name__, self.pk)

    def to_dict(self):
        return {
            **model_to_dict(self),
            "global_id": self.global_id,
        }

    def save(self, *args, **kwargs):
        self.full_clean()
        return super().save(*args, **kwargs)



from django.contrib.auth.models import AbstractUser
from django.db import models
from django.utils.translation import gettext_lazy as _

from .base import BaseORMModel


class User(AbstractUser, BaseORMModel):
    """
    사용자 모델
    """

    uid = models.CharField(
        _("Firebase UID"),
        primary_key=True,
        max_length=128,
        null=False,
        blank=False,
        editable=False,
    )

    email = models.EmailField(
        _("이메일 주소"),
        null=False,
        blank=False,
    )
    email_verified = models.BooleanField(
        _("이메일 인증 여부"),
        default=False,
    )
    provider_data = models.JSONField(
        _("제공 업체 데이터"),
        null=True,
        blank=True,
    )

    first_name = None
    last_name = None

    REQUIRED_FIELDS = ["email", "uid"]

    class Meta:
        db_table = "user"
        verbose_name = _("사용자")
        verbose_name_plural = _("사용자")
        indexes = [models.Index(fields=["email"], name="email_index")]

    def __str__(self):
        return f"<User: {self.email}>"



from datetime import datetime, timezone

from asgiref.sync import sync_to_async
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.hashers import make_password
from django.contrib.auth.models import AbstractBaseUser, AnonymousUser
from firebase_admin import auth
from firebase_admin.auth import UserInfo, UserRecord

AUTHORIZATION_HEADER = getattr(
    settings,
    "AUTHORIZATION_HEADER",
    "Authorization",
)
AUTHORIZATION_HEADER_PREFIX = getattr(
    settings,
    "AUTHORIZATION_HEADER_PREFIX",
    "Bearer",
)
COOKIE_NAME = getattr(settings, "COOKIE_NAME", None)
AUTO_CREATE_USER = getattr(settings, "AUTO_CREATE_USER", False)


UserModel = get_user_model()


def is_valid_jwt(token):
    return len(token.split(".")) == 3


def timestamp_to_datetime(timestamp: int) -> datetime:
    return datetime.fromtimestamp(timestamp / 1000, tz=timezone.utc)


def provider_data_to_dict(provider_data: list[UserInfo]) -> dict:
    return [
        {
            "uid": p.uid,
            "email": p.email,
            "display_name": p.display_name,
            "phone_number": p.phone_number,
            "photo_url": p.photo_url,
            "provider_id": p.provider_id,
        }
        for p in provider_data
    ]


def create_user(user_record: UserRecord) -> AbstractBaseUser | None:
    if not AUTO_CREATE_USER:
        return None

    user, _ = UserModel.objects.update_or_create(
        uid=user_record.uid,
        defaults={
            "username": user_record.email,
            "password": make_password(f"{user_record.provider_id}:{user_record.uid}"),
            "email": user_record.email,
            "email_verified": user_record.email_verified,
            "date_joined": timestamp_to_datetime(user_record.user_metadata.creation_timestamp),
            "last_login": timestamp_to_datetime(user_record.user_metadata.last_sign_in_timestamp),
            "is_superuser": False,
            "is_staff": False,
            "provider_data": provider_data_to_dict(user_record.provider_data),
        },
    )

    return user


def get_user(request):
    # request에서 토큰 가져오기
    token = request.headers.get(AUTHORIZATION_HEADER)
    if COOKIE_NAME:
        token = token or request.COOKIES.get(COOKIE_NAME)

    if not token:
        return

    # 토큰 형식 검증
    token = token.replace(AUTHORIZATION_HEADER_PREFIX, "").strip()
    if not is_valid_jwt(token):
        return

    # 토큰 검증
    decoded_token = auth.verify_id_token(token)

    # Firebase에서 사용자 정보 조회
    user_record = auth.get_user(decoded_token["uid"])

    # DB에 사용자 정보 조회 또는 생성
    user = None
    try:
        # 사용자 정보 조회
        user = UserModel.objects.get(uid=user_record.uid)

        # 변경이 필요한 필드만 업데이트
        updates = {}

        if user.email_verified != user_record.email_verified:
            updates["email_verified"] = user_record.email_verified

        npd = provider_data_to_dict(user_record.provider_data)
        if user.provider_data != npd:
            updates["provider_data"] = npd

        nll = timestamp_to_datetime(user_record.user_metadata.last_sign_in_timestamp)
        if user.last_login != nll:
            updates["last_login"] = nll

        # 변경사항이 있을 때만 update() 호출
        if updates:
            UserModel.objects.filter(uid=user.uid).update(**updates)

        user.refresh_from_db()
    except UserModel.DoesNotExist:
        user = create_user(user_record)

    return user or AnonymousUser()


async def aget_user(request):
    """See get_user()."""
    return await sync_to_async(get_user)(request)



# Generated by Django 5.1.5 on 2025-02-05 07:58

import django.contrib.auth.models
import django.contrib.auth.validators
import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("auth", "0012_alter_user_first_name_max_length"),
    ]

    operations = [
        migrations.CreateModel(
            name="User",
            fields=[
                ("password", models.CharField(max_length=128, verbose_name="password")),
                (
                    "last_login",
                    models.DateTimeField(blank=True, null=True, verbose_name="last login"),
                ),
                (
                    "is_superuser",
                    models.BooleanField(
                        default=False,
                        help_text="Designates that this user has all permissions without explicitly assigning them.",
                        verbose_name="superuser status",
                    ),
                ),
                (
                    "username",
                    models.CharField(
                        error_messages={"unique": "A user with that username already exists."},
                        help_text="Required. 150 characters or fewer. Letters, digits and @/./+/-/_ only.",
                        max_length=150,
                        unique=True,
                        validators=[django.contrib.auth.validators.UnicodeUsernameValidator()],
                        verbose_name="username",
                    ),
                ),
                (
                    "is_staff",
                    models.BooleanField(
                        default=False,
                        help_text="Designates whether the user can log into this admin site.",
                        verbose_name="staff status",
                    ),
                ),
                (
                    "is_active",
                    models.BooleanField(
                        default=True,
                        help_text="Designates whether this user should be treated as active. Unselect this instead of deleting accounts.",
                        verbose_name="active",
                    ),
                ),
                (
                    "date_joined",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="date joined"
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True, help_text="생성일")),
                ("modified_at", models.DateTimeField(auto_now=True, help_text="수정일")),
                (
                    "uid",
                    models.CharField(
                        editable=False,
                        max_length=128,
                        primary_key=True,
                        serialize=False,
                        verbose_name="Firebase UID",
                    ),
                ),
                ("email", models.EmailField(max_length=254, verbose_name="이메일 주소")),
                (
                    "email_verified",
                    models.BooleanField(default=False, verbose_name="이메일 인증 여부"),
                ),
                (
                    "provider_data",
                    models.JSONField(blank=True, null=True, verbose_name="제공 업체 데이터"),
                ),
                (
                    "groups",
                    models.ManyToManyField(
                        blank=True,
                        help_text="The groups this user belongs to. A user will get all permissions granted to each of their groups.",
                        related_name="user_set",
                        related_query_name="user",
                        to="auth.group",
                        verbose_name="groups",
                    ),
                ),
                (
                    "user_permissions",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Specific permissions for this user.",
                        related_name="user_set",
                        related_query_name="user",
                        to="auth.permission",
                        verbose_name="user permissions",
                    ),
                ),
            ],
            options={
                "verbose_name": "사용자",
                "verbose_name_plural": "사용자",
                "db_table": "user",
                "indexes": [models.Index(fields=["email"], name="email_index")],
            },
            managers=[
                ("objects", django.contrib.auth.models.UserManager()),
            ],
        ),
    ]



from django.apps import AppConfig


class DjangoFirebaseAuthConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "django_firebase_auth"
    label = "django_firebase_auth"
